/* GDK - The GIMP Drawing Kit
 *
 * Copyright (C) 2017 Benjamin Otte <otte@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __GDK_PIPE_IO_STREAM_H__
#define __GDK_PIPE_IO_STREAM_H__

#include <gdk/gdkversionmacros.h>
#include <gio/gio.h>

G_BEGIN_DECLS


GIOStream *             gdk_pipe_io_stream_new                          (void);


G_END_DECLS

#endif /* __GDK_PIPE_IO_STREAM_H__ */
